/*
 disasm68k.c
  Disassembler for 680x0 series, including 68060.

   Author: Adam Pickett
  Created: 2021.09
  Revised: 2021.10.18

2021.09.23
  moved to separate source file from reveng.c.
*/

#include "enviro64.h"
#include "ftrig.h"
//#include <stdio.h>
//#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "disasm68k.h"

/* ---- library */

#ifdef __POCC__
 #pragma warn(disable:2237)
#endif

const char	ubasechars[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_";

/* ---- code */

static const char	op_name[][6] =
{
	"dc", "dc",
	"dc",
	"trap",
	"rtm",
	"exg",
	"ext", "extb",
	"swap",
	"moveq",
	"pack", "unpk",
	"addx", "subx",
	"abcd", "sbcd",

	"asr", "asl",
	"lsr", "lsl",
	"roxr", "roxl",
	"ror", "rol",
	"ori", "andi",
	"subi", "addi",
	"eori",
	"cmpi",
	"cas","cas2",

	"nbcd",
	"eor",
	"mulu", "muls",
	"divu", "divs",
	"divul", "divsl",
	"btst",
	"bchg",
	"bclr",
	"bset",
	"tas",
	"move",
	"movem",
	"move16",

	"fmovem",
	"fnop",
	"fb\362", "fs\342",
	"fsave", "frest\200",
	"cinva", "cpusha",
	"pmove",
	"pflush",
	"pb\321",
	"ps\301",
	"psave", "prest\200",
	"jsr", "jmp",

	"bra", "bsr",
	"b\330", "s\330",
	"movea",
	"moves",
	"cmpa", "cmp",
	"cmp2",
	"tst",
	"addq", "subq",
	"adda", "suba",
	"add", "sub",

	"chk2", "cmpm",
	"clr",
	"not",
	"negx", "neg",
	"or",
	"and",
	"chk",
	"callm",
	"bkpt",
	"db\330",
	"t\330", "tp\330",
	"movec",
	"movep",

	"bftst", "bfextu",
	"bfchg", "bfexts",
	"bfclr", "bfffo",
	"bfset", "bfins",
	"link", "unlk",
	"lea",
	"pea",
	"reset",
	"nop",
	"stop",
	"rte",

	"rtd", "rts",
	"trapv",
	"rtr",
	"ptestw", "ptestr",
	"ploadw", "ploadr",
	"plpaw", "plpar",
	"pflus\201",
	"pflus\202",
	"pflus\203",
	"pflus\204",
	"pflus\205",
	"pmove\206",

	"pvalid",
	"pdb\301",
	"ptrap\301",
	"cinvl", "cpushl",
	"cinvp", "cpushp",
	"bgnd",
	"tblu", "tblun",
	"tbls", "tblsn",
	"lpstop",
	"db\342",
	/* fxx */
	"trap\342",
	"movecr",
	"move",
	"int",
	"sinh",
	"intrz",
	"sqrt",
	"lognp1", "etoxm1",
	"tanh",
	"atan",
	"asin",
	"atanh",
	"sin",
	"tan",
	"etox", "twotox", "tentox",
	"logn", "log10", "log2",
	"abs",
	"cosh",
	"neg",
	"acos",
	"cos",
	"getexp", "getman",
	"div", "mod",
	"add",
	"mul",
	"sgldiv",
	"rem",
	"scale",
	"sgldiv",
	"sub",
	"sincos",
	"cmp",
	"test",
	"smove", "ssqrt",
	"dmove", "dsqrt",
	"sabs", "sneg",
	"dabs", "dneg",
	"sdiv", "sadd", "smul",
	"ddiv", "dadd", "dmul",
	"ssub", "dsub"
};

static const uint8 fpopx[128] =
{
	144,145,146,147,148,0,149,0,
	150,151,152,0,153,0,154,155,
	156,157,158,0,159,160,161,0,
	162,163,164,0,165,166,167,168,
	169,0,170,171,172,173,174,175,
	176,0,0,0,0,0,0,0,
	177,177,177,177,177,177,177,177,
	178,0,179,0,0,0,0,0,
	180,181,0,0,182,183,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	184,0,185,0,186,0,187,0,
	188,0,189,190,191,0,192,193,
	194,0,0,0,195
};

static const char	fpcrx[64] =
{
	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 4, 5, 6,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22
};

static const char	fpcrn[][6] =
{
	"Pi",
	"g10(2)",
	"e",
	"g2(e)",
	"g10(e)",
	"0", /* all values default to 0 though */
	"ln(2)",
	"ln(10)",
	"100",
	"+1",
	"+2",
	"+4",
	"+8",
	"+16",
	"+32",
	"+64",
	"+128",
	"+256",
	"+512",
	"+1024",
	"+2048",
	"+4096"
};

static const char	cc_name[71][3] =
{
	"t","f","hi","ls",
	"cc","cs","ne","eq",
	"vc","vs","pl","mi",
	"ge","lt","gt","le",

	"bs","bc","ls","lc",
	"ss","sc","as","ac",
	"ws","wc","is","ic",
	"gs","gc","cs","cc",

	"f","eq","ogt","oge",
	"olt","ole","ogl","or",
	"un","ueq","ugt","uge",
	"ult","ule","ne","t",
	"sf","seq","gt","ge",
	"lt","le","gl","gle",
	"\0le","ngl","nle","nlt",
	"\0e","ngt","sne","st",
	/* name extensions for longer than 6 chars */
	"ore", "hn", "han", "ha",
	"hr", "hs", "fd"
};

static char *strcatop(char *d,
	uint8 o,
	uint32 op)
{
	const char	*f = op_name[o];
	char		c;

	if (o>=_op_fdbcc)
		*d++ = 'f';
	o = 6, c = *f++;
	do
	{
		if (c<0)
		{
			f = cc_name[(o=(c>>5)+4)==0 ? (c&31)+64 /* name extension */
				:(op>>(c&24)&~((uint32)~3<<o))+(c<<4&0x70)]; /* condition ext. */
			o = 3;
			if ((c=*f++)==0)
				*d++ = 'n', c = 'g';
		}
		*d++ = c;
	} while(--o && (c=*f++)!=0);
	return d;
}

enum
{ /* operand addressing modes */
	_am_implied,
	_am_dn,	/* Dn		;data register */
	_am_an,	/* An 		;address reg. */
	_am_ai,	/* (An)		;indirect */
	_am_aip,/* (An)+	;indir., post-increment */
	_am_mai,/* -(An)	;indir., pre-decr. */
	_am_dai,/* W(An)	;indir., displaced */
	_am_aix,/*(n,An[,Xn]..) ;indir., indexed/complex */

	_am_abw,/* W		;8 - absolute indir. */
	_am_abl,/* L */
	_am_rel,/* *+W		;relative indir. */
	_am_rex,/* (*+[,Xn]..)	;rel. indexed/complex */
	_am_imz,/* #oz		;op1+ immediate, size dep. */
	_am_bf,	/* ea{..}	;bitfield or datafield in preprocessing */
	_am_ccr,/* CCR		;condition codes reg. */
	_am_sr,	/* SR		;status register - "implied" for mmode */

	_am_dh,	/* Dh1		;16 - high register of op1 */
	_am_rn,	/* Rn1		;high 4 bits of op1, Dn|An; ex: MOVES */
	_am_rez,/* *+oz		;branch offset */
	_am_dcw,/* $W		;raw */
	_am_sq,	/* #s8		;op0, low 8-bits signed */
	_am_hq,	/* #h0		;op0, bits 9-11, 0=8 */
	_am_dlm,/* Dl,Dm	;middle and low register for CAS */
	_am_rnl,/*..[Dn]-/[An]..;register list */

	_am_dlh,/* Dh1:Dl1	;24 - high/low regs. for MULx/DIVx */
	_am_dlc,/* Dl1:Dl2	;low op1/2 Dc for CAS2 */
	_am_dmu,/* Dm1:Dm2	;middle op1/2 Du for CAS2 */
	_am_rnc,/* (Rn1):(Rn2)	;regs. for CAS2 - can be (Dn) indirect */
	_am_isz,/* #oz		;signed immediate */
	_am_dmn,/* Dn:Dl1	;regs. for TBLxx */
	_am_fpn,/* FPn		;FPU gp reg. (all FPn are in op1) */
	_am_fph,/* FPh		;high FPU reg. */

	_am_fpl,/* FPl		;32 - low FPU reg. for FSINCOS */
	_am_fnl,/*..[FPn]/..	;FPn list */
	_am_fnd,/* Dm		;FPn dynamic list */
	_am_fcl,/*..[FPxr]/..	;FP control reg. list */
	_am_dc3,/* #imm3	;low 3 bits of op1 */
	_am_pta,/* #h1[,An]	;PTEST level & optional An last desc. */
	_am_pfk,/* #mask	;PFLUSH mask */
	_am_al,	/* Al1		;low 3 bits of op1, An */

	_am_val,/* VAL		;40 - valid access level */
	_am_tt0,/* AC|TT0	;MMU transparent translation regs. */
	_am_tt1,
	_am_drp,/* DRP		;DMA root pointer */
	/* cache modes */
	_am_nc,	/* none */
	_am_dc,	/* data */
	_am_ic,	/* instruction */
	_am_dic,/* data & ins. */
	/* MOVEC control registers */
	_am_sfc,		/* 48 - Source Function Code */
	_am_dfc,/* Dest Func. Code */
	_am_cacr,/* cache control reg. */
	_am_tc,
	/* 040 only; on EC040, these 4 are access control regs. */
	_am_itt0,/* Ins. Transparent Translation Reg. 0 */
	_am_itt1,
	_am_dtt0,/* Data Transparent Translation Reg. 0 */
	_am_dtt1,

	_am_usp,		/* 56 - user stack pointer */
	_am_vbr,/* vector base reg. */
	_am_caar,/* cache address reg. */
	_am_msp,/* master stack pointer */
	_am_isp,/* isp stack pointer */
	_am_mmusr,/* MMU status reg. (040 only) */
	_am_urp,/* user root pointer (040 only) */
	_am_srp,/* supervisor root pointer (040 only) */

	_am_crp,		/* 64 - CPU root pointer */
	_am_cal,/* current access level */
	_am_scc,/* stack change control */
	_am_ac,	/* access control */
	_am_psr,/* PMMU SR */
	_am_pcsr,/* PMMU cache SR - read-only */
	_am_badx,/* bkpt ack data #reg */
	_am_bacx,/* bkpt ack control #reg */
	/* '060 */
	_am_buscr,/* bus control reg. */
	_am_pcr,/* process config. reg. */

	_am_count
};

static const char	am_strs[_am_count][5] =
{
	"",
	"dN",
	"aN",
	"(aN)",
	"(aN)+",
	"-(aN)",
	"W(aN)",
	"XSEGT",
	"W",
	"P",
	"Y",
	"XZERT",
	"#U",
	"{K}",
	"ccr",
	"sr",

	"D\xC",
	"R",
	"Z",
	"O",
	"#Q",
	"#H",
	"D\0,D\6",
	"L",

	"D\0:D\xC",
	"D\0:J\0",
	"D\6:J\6",
	"(R)I)",
	"#S",
	"dN:D\0",
	"fpV7",
	"fpV\xA",

	"fpV\0",
	"M",
	"F",
	"D\4",
	"#V",
	"#V\tC",
	"#V\4",
	"aV",

	"val",
	"tt0",
	"tt1",
	"drp",

	"nc",
	"dc",
	"ic",
	"dc/ic",

	"sfc",
	"dfc",
	"cacr",
	"tc",
	"itt0",
	"itt1",
	"dtt0",
	"dtt1",
	"usp",
	"vbr",
	"caar",
	"msp",
	"isp",
	"mmusr",
	"urp",
	"srp",

	"crp",
	"cal",
	"scc",
	"ac",
	"psr",
	"pcsr",
	"badV\2",
	"bacV\2",

	"buscr",
	"pcr"
};

static char *catfparm(char *d,
	short v)
{
	short		q;

	if (v&32)
		v&= 7, *d++ = 'd';
	else
	{
		v&= 31;
		if ((q=v*13>>7)!=0)
			*d++ = '0'+(char)q, q*= 10, v-= q;
	}
	v+= '0', *d++ = (char)v;
	return d;
}

static char *catmparm(char *d,
	uint8 t,
	uint8 m,
	char n)
{
	char		i = 0;

	if (m && t)
		*d++ = '/', m = 0; /* continuing from prev. list */
	do
		if (t&1)
		{
			if (m && t&2)
				continue;
			if (m++)
				*d++ = m>1 ? '-':'/';
			*d++ = n;
			if (n=='f')
				*d++ = 'p';
			*d++ = '0'+i;
		} else
		{
			if (m && t)
				*d++ = '/';
			m = 0;
		}
	while(t>>=1,++i<8);
	return d;
}

static char *strcatam(char *d,
	int m,
	int32 op,
	const uint8 *s)
{
	uint8		x[4]; /* displacement signature, Areg byte, complex opword */
	const char	*f = am_strs[m], *e = f+5;
	int32		t;
	char		c;

	for(*(int32*)x=0; (c=*f++)!=0;)
	{
		switch(c)
		{
		case 'C': /* optional descriptor Rn for PTEST */
			if (!(op&256))
				break;
			*d++ = ',', t = s[1]>>4;
			goto rn;
		case 'X': /* indexed/complex primer */
			*x = *f, x[1] = (uint8)op;
			*(int16*)(x+2) = *((const int16*)s)++;
			t = x[3], op = t|(128<<16);
			if (x[2]&1)
			{ /* complex op1.b not an offset, nor does value appear before '(' */
				op =-1<<16; /* value for offset defaults to zero */
				if ((t&=0x30)>31)
					op = (t-_oz_w+2)<<16-4;
				*d++ = '(';
				c = '[';
				if (x[3]&7)
					goto ch;
				break;
			}
			break;
		case 'T': /* complex closer */
			c = x[3]&3;
			if (x[2]&1 && (op=c+0u<<16)!=0)
			{
				if (!(x[3]&4))
					*d++ = ']';
				if (c>1)
					{f = ",S)", e = f+4; break;}
			}
			c = ')';
			goto ch;
		case 'E': /* first indexed separator */
			c = '(';
			if (x[2]&1)
			{
				if ((x[3]|((*x=='Z')<<7))>191) /* no index or An */
				{
					++f;
					if (x[3]&4) /* indirect before close */
						*d++ = ']';
					break;
				}
				if (*x=='Z')
					break;
				c = ','; /* has index and/or An */
			}
			goto ch;
		case 'G': /* An[,Rn] */
			if (x[2]&1 && x[3]&0x80)
			{
				//if (x[3]>191) /* no registers */
				//	break;
				*d++ = 'z'; /* redundant no-An complex mode, but eh */
			}
			*d++ = 'a', *d++ = (x[1]&7)+'0';
			/* fall thru */
		case 'R': /* Rn */
			t = *s;
			if (*x)
			{
				if ((uint8)((t=x[2])&1))
				{
					if (x[3]&64) /* no index */
						break;
					if (x[3]&4) /* indirect before index */
						*d++ = ']';
					*d++ = ',';
				} else if (*x!='Z')
					*d++ = ',';
			}
			t>>= 4;
		rn:	if (!(char)(t&8))
				goto dn;
			*d++ = 'a';
			goto n3;
		case 'J': /* op2 data reg */
			t = be16_native(((const uint16*)s)[1])>>(*f++); goto dn;
		case 'D': /* op1 data reg */
			t = be16_native(*(const uint16*)s)>>(*f++);
		dn:	*d++ = 'd';
			goto n3;
		case 'V': /* 3/4-bit value/reg no. within op1 */
			c = *f++;
			t = be16_native(*(const uint16*)s)>>c;
			if (c<0)
				{t&= 15; goto uv;}
			goto n3;
		case 'N': /* 3-bit reg number in op */
			t = op;
		n3:	*d++ = (char)((t&7)+'0');
			if ((c=x[2]>>1)==0)
				break;
			if (c&4)
				*((uint16*)d)++ = MK2CC('.','l');
			if (c&=3)
				*d++ = '*', *d++ = '0'+(1u<<c);
			break;
		case 'Z': /* rel. displacment, op-size */
			if ((int8)(t=op>>16)<=_oz_implied)
			{
				t = (int8)op;
				if (op<0)
					{c = 0; goto sv;}
				goto rv;
			}
			if ((uint8)t==_oz_l)
				t = be32_native(*((const int32*)s)++);
			//else if ((uint8)t==_oz_b)
			//	t = ((int8*)s)[1];
			else
		case 'Y':	t = (int16)be16_native(*((const int16*)s)++);
		rv:	*d++ = '*';
			if (!(t+=2))
				break;
			if (t<0)
				t =-t, *d++ = '-';
			else
				*d++ = '+';
			goto uv;
		case 'U': /* unsigned/unknown sign immediate */
			if ((op&=-1<<16)!=(_oz_l<<16))
			{ /* implied is 16-bit */
				t = s[1];
				if (op!=(_oz_b<<16))
				#if BIG_ENDIAN
					t = *(const uint16*)s;
				#else
					t|= *s<<8;
				#endif
				goto uv;
			}
		case 'P': /* raw long/pointer */
			t = be32_native(*(const int32*)s), m = 32;
			goto hex;
		case 'H': /* quick op3 */
			t = (((op>>9)-1)&7)+1;
			goto uv;
		case 'Q': goto sq; /* can't move this case to another switch */
		case 'S': /* signed imm */
			m = 32;
			switch(op>>16)
			{
			default: sq: t = (int8)op; break; /* quick op8, signed */
			case _oz_b: t = ((int8*)s)[1]; break;
			case _oz_implied:
			case _oz_w: t = (int16)be16_native(*((const int16*)s)++); break;
			case _oz_l: t = be32_native(*((const int32*)s)++); break;
			case _oz_c: t = s[1]&63; goto uv;
			case _oz_s: goto fhx;
			case _oz_x: case _oz_p: m+= 256;
			case _oz_d: m+= 256;
			fhx:	*((int16*)d)++ = MK2CC('$','"'), f = "\"", e = f+4;
				t = be32_native(*(const int32*)s);
				goto hxd;
			}
		sv:	if (t<0)
				t =-t, *d++ = '-';
		uv:	m = 32-(cntlz32(t)&(28|32));
			if (t>9 || c>='Y') /* some readers depend on the $ in relatives */
				goto hex;
			goto hxd;
		case 'O': /* raw opcode */
			t = op;
			goto h16;
		case 'W': /* word address/displacement, signed */
			if ((t=(int16)be16_native(*(const int16*)s))<0)
				t =-t, *d++ = '-';
		h16:	m = 16;
		hex:	*d++ = '$';
		hxd:	for(;;s+=4,t=be32_native(*(const int32*)s),m+=32-256)
			{
				do
					m-= 4, *d++ = ubasechars[(uint32)t>>m&15];
				while((int8)m>0);
				if (m<=0)
					break;
			}
			break;
		case 'K': /* D|#:D|# bitfield */
			t = be16_native(*(const uint16*)s);
			d = catfparm(d,t>>6);
			*d++ = ':';
			d = catfparm(d,t);
			break;
		case 'L': /* Rn list */
			t = be16_native(*(const uint16*)s);
			if ((uint8)(op&0x38)==32) /* -(An) */
			{
				m = t, t = 0, c = 16;
				do
					t+= t+(m&1);
				while(m>>=1,--c);
			}
			d = catmparm(d,(uint8)t,0,'d');
			d = catmparm(d,(uint8)(t>>8),(uint8)t,'a');
			break;
		case 'M': /* FPn list */
			d = catmparm(d,be16_native(*(const uint16*)s),0,'f');
			break;
		case 'F': /* FPxr list */
			c = *s, m = 0;
			if (c&1<<4)
				m = 1, *((int32*)d)++ = MK4CC('f','p','c','r');
			if (c&1<<3)
			{
				if (m)
					*d++ = '/';
				m = 1, *((int32*)d)++ = MK4CC('f','p','s','r');
			}
			if (!(c&1<<2))
				break;
			if (m)
				*d++ = '/';
			*((int32*)d)++ = MK4CC('f','p','i','a'), c = 'r';

		default:
		ch:	*d++ = c;
		}
		if (f>=e)
			break;
	}
	return d;
}

static int strcomam(char *b,
	int m,
	int32 op,
	const uint8 *s,
	int offs)
{ /* really only have to handle relative, imm, sq and fmovecr */
	const char	*f = am_strs[m], *e = f+5;
	char		*d = b;
	int32		t = op>>16;
	uint8		c;

	*((int16*)d)++ = MK2CC(';',' ');
	for(m=0; (c=*f++)!=0;)
	{
		switch(c)
		{
		case 'X': m = be16_native(*((int16*)s++));
			if (!(uint16)(m&0x100))
				op = m, t = 0;
			else if (!(uint8)((t=m>>4&3)&2))
				return 0;
			break;
		case 'S':
			if ((uint8)t!=_oz_c || (t=fpcrx[s[1]&63])==0)
				return 0;
			s = (const uint8*)fpcrn[t-1];
			if ((c=*s++)=='+')
				*((int16*)d)++ = MK2CC('1','e');
			else if (c=='g')
				*((int16*)d)++ = MK2CC('L','o');
			m = 6;
			do
				*d++ = c;
			while((c=*s++)!=0 && --m);
			goto end;
		case 'Q': t = (uint8)op; goto lit;
		case 'U': t = t==_oz_l ? be32_native(*(uint32*)s)
				:be16_native(*(uint16*)s);
		lit:	if ((m=32-(cntlz32(t)&(24|32)))==0)
				return 0;
			*d++ = '\'';
			do
				if (m-=8,(uint8)((c=(uint32)t>>m)-9u)<14u-9)
					*d++ = '\\', *d++ = "tnvfr"[c-9];
				else if (c=='?')
					*((int16*)d)++ = MK2CC('\\','?'); /* avoid trigraphs */
				else if (c-' '+0u>126-' ')
					return 0;
				else
				{
					if (c=='\\' || c=='\'')
						*d = '\\';
					*d++ = c;
				}
			while(m>0);
			*d++ = '\'';
			goto end;
		case 'Z':
			if (t==_oz_implied)
				t = (int8)op;
			else if ((uint8)t==_oz_l)
				t = (int32)be32_native(*(uint32*)s);
			else
		case 'Y':	t = (int16)be16_native(*(uint16*)s);
			t+= 2, t+= offs, m = 32; //t+0u>0xFFFFFF ? 32:24;
			do
				m-= 4, *d++ = ubasechars[(uint32)t>>m&15];
			while((int8)m>0);
			goto end;
		}
		if (f>=e)
			break;
	}
	*d = 0;
end:	return (t=d-b)==2 ? 0:t;
}

static const unsigned short	opinfo[0x10000>>6] =
{ /* mmode.3 opix.7 size.2 dmode.4 */
	0x061C,0x062C,0x063C,0x121D,0x0A01,0x0A41,0x0A81,0x0AC1,/*0000*/
	0x065C,0x066C,0x067C,0x122D,0x0A01,0x0A41,0x0A81,0x0AC1,/*0200*/
	0x069C,0x06AC,0x06BC,0x123D,0x0A01,0x0A41,0x0A81,0x0AC1,/*0400*/
	0x06DC,0x06EC,0x06FC,0x0100,0x0A01,0x0A41,0x0A81,0x0AC1,/*0600*/
	0x0A0C,0x0A4C,0x0A8C,0x0ACC,0x0A01,0x0A41,0x0A81,0x0AC1,/*0800*/
	0x071C,0x072C,0x073C,0xE79D,0x0A01,0x0A41,0x0A81,0x0AC1,/*0A00*/
	0x075C,0x076C,0x077C,0xE7AD,0x0A01,0x0A41,0x0A81,0x0AC1,/*0C00*/
	0x115D,0x116D,0x117D,0xE7BD,0x0A01,0x0A41,0x0A81,0x0AC1,/*0E00*/
	0x0B51,0x0B52,0x0B53,0x0B54,0x0B55,0x0B56,0x0B57,0x0B58,/*1000*/
	0x0B51,0x0B52,0x0B53,0x0B54,0x0B55,0x0B56,0x0B57,0x0B59,/*1200*/
	0x0B51,0x0B52,0x0B53,0x0B54,0x0B55,0x0B56,0x0B57,0x0080,/*1400*/
	0x0B51,0x0B52,0x0B53,0x0B54,0x0B55,0x0B56,0x0B57,0x0080,/*1600*/
	0x0B51,0x0B52,0x0B53,0x0B54,0x0B55,0x0B56,0x0B57,0x0080,/*1800*/
	0x0B51,0x0B52,0x0B53,0x0B54,0x0B55,0x0B56,0x0B57,0x0080,/*1A00*/
	0x0B51,0x0B52,0x0B53,0x0B54,0x0B55,0x0B56,0x0B57,0x0080,/*1C00*/
	0x0B51,0x0B52,0x0B53,0x0B54,0x0B55,0x0B56,0x0B57,0x0080,/*1E00*/
	0x0B71,0x1132,0x0B73,0x0B74,0x0B75,0x0B76,0x0B77,0x0B78,/*2000*/
	0x0B71,0x1132,0x0B73,0x0B74,0x0B75,0x0B76,0x0B77,0x0B79,/*2200*/
	0x0B71,0x1132,0x0B73,0x0B74,0x0B75,0x0B76,0x0B77,0x0080,/*2400*/
	0x0B71,0x1132,0x0B73,0x0B74,0x0B75,0x0B76,0x0B77,0x0080,/*2600*/
	0x0B71,0x1132,0x0B73,0x0B74,0x0B75,0x0B76,0x0B77,0x0080,/*2800*/
	0x0B71,0x1132,0x0B73,0x0B74,0x0B75,0x0B76,0x0B77,0x0080,/*2A00*/
	0x0B71,0x1132,0x0B73,0x0B74,0x0B75,0x0B76,0x0B77,0x0080,/*2C00*/
	0x0B71,0x1132,0x0B73,0x0B74,0x0B75,0x0B76,0x0B77,0x0080,/*2E00*/
	0x0B61,0x1122,0x0B63,0x0B64,0x0B65,0x0B66,0x0B67,0x0B68,/*3000*/
	0x0B61,0x1122,0x0B63,0x0B64,0x0B65,0x0B66,0x0B67,0x0B69,/*3200*/
	0x0B61,0x1122,0x0B63,0x0B64,0x0B65,0x0B66,0x0B67,0x0080,/*3400*/
	0x0B61,0x1122,0x0B63,0x0B64,0x0B65,0x0B66,0x0B67,0x0080,/*3600*/
	0x0B61,0x1122,0x0B63,0x0B64,0x0B65,0x0B66,0x0B67,0x0080,/*3800*/
	0x0B61,0x1122,0x0B63,0x0B64,0x0B65,0x0B66,0x0B67,0x0080,/*3A00*/
	0x0B61,0x1122,0x0B63,0x0B64,0x0B65,0x0B66,0x0B67,0x0080,/*3C00*/
	0x0B61,0x1122,0x0B63,0x0B64,0x0B65,0x0B66,0x0B67,0x0080,/*3E00*/
	0x1510,0x1520,0x1530,0x07CF,0x1631,0x0080,0x1621,0x01C0,/*4000*/
	0x1490,0x14A0,0x14B0,0x07CE,0x1631,0x0080,0x1621,0x01C0,/*4200*/
	0x1550,0x1560,0x1570,0x0B4E,0x1631,0x0080,0x1621,0x01C0,/*4400*/
	0x14D0,0x14E0,0x14F0,0x0B4F,0x1631,0x0080,0x1621,0x01C0,/*4600*/
	0x0800,0x0200,0x01A0,0x01B0,0x1631,0x0080,0x1621,0x01F1,/*4800*/
	0x1250,0x1260,0x1270,0x0B00,0x1631,0x0080,0x1621,0x01C0,/*4A00*/
	0x09FD,0x09BD,0x0BAD,0x0BBD,0x1631,0x0080,0x1621,0x01C0,/*4C00*/
	0x0080,0x00C0,0x0F80,0x0FC0,0x1631,0x0080,0x1621,0x01C0,/*4E00*/
	0xD290,0xD2A0,0xD2B0,0x10C0,0xD2D0,0xD2E0,0xD2F0,0x10C0,/*5000*/
	0xD290,0xD2A0,0xD2B0,0x10C0,0xD2D0,0xD2E0,0xD2F0,0x10C0,/*5200*/
	0xD290,0xD2A0,0xD2B0,0x10C0,0xD2D0,0xD2E0,0xD2F0,0x10C0,/*5400*/
	0xD290,0xD2A0,0xD2B0,0x10C0,0xD2D0,0xD2E0,0xD2F0,0x10C0,/*5600*/
	0xD290,0xD2A0,0xD2B0,0x10C0,0xD2D0,0xD2E0,0xD2F0,0x10C0,/*5800*/
	0xD290,0xD2A0,0xD2B0,0x10C0,0xD2D0,0xD2E0,0xD2F0,0x10C0,/*5A00*/
	0xD290,0xD2A0,0xD2B0,0x10C0,0xD2D0,0xD2E0,0xD2F0,0x10C0,/*5C00*/
	0xD290,0xD2A0,0xD2B0,0x10C0,0xD2D0,0xD2E0,0xD2F0,0x10C0,/*5E00*/
	0x7000,0x7000,0x7000,0x7000,0x7040,0x7040,0x7040,0x7040,/*6000*/
	0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,/*6200*/
	0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,/*6400*/
	0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,/*6600*/
	0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,/*6800*/
	0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,/*6A00*/
	0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,/*6C00*/
	0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,0x7080,/*6E00*/
	0xA241,0xA241,0xA241,0xA241,0x0080,0x0080,0x0080,0x0080,/*7000*/
	0xA241,0xA241,0xA241,0xA241,0x0080,0x0080,0x0080,0x0080,/*7200*/
	0xA241,0xA241,0xA241,0xA241,0x0080,0x0080,0x0080,0x0080,/*7400*/
	0xA241,0xA241,0xA241,0xA241,0x0080,0x0080,0x0080,0x0080,/*7600*/
	0xA241,0xA241,0xA241,0xA241,0x0080,0x0080,0x0080,0x0080,/*7800*/
	0xA241,0xA241,0xA241,0xA241,0x0080,0x0080,0x0080,0x0080,/*7A00*/
	0xA241,0xA241,0xA241,0xA241,0x0080,0x0080,0x0080,0x0080,/*7C00*/
	0xA241,0xA241,0xA241,0xA241,0x0080,0x0080,0x0080,0x0080,/*7E00*/
	0x1591,0x15A1,0x15B1,0x0921,0x03C1,0x0281,0x02C1,0x0961,/*8000*/
	0x1591,0x15A1,0x15B1,0x0921,0x03C1,0x0281,0x02C1,0x0961,/*8200*/
	0x1591,0x15A1,0x15B1,0x0921,0x03C1,0x0281,0x02C1,0x0961,/*8400*/
	0x1591,0x15A1,0x15B1,0x0921,0x03C1,0x0281,0x02C1,0x0961,/*8600*/
	0x1591,0x15A1,0x15B1,0x0921,0x03C1,0x0281,0x02C1,0x0961,/*8800*/
	0x1591,0x15A1,0x15B1,0x0921,0x03C1,0x0281,0x02C1,0x0961,/*8A00*/
	0x1591,0x15A1,0x15B1,0x0921,0x03C1,0x0281,0x02C1,0x0961,/*8C00*/
	0x1591,0x15A1,0x15B1,0x0921,0x03C1,0x0281,0x02C1,0x0961,/*8E00*/
	0x13D1,0x13E1,0x13F1,0x1362,0x0351,0x0361,0x0371,0x1372,/*9000*/
	0x13D1,0x13E1,0x13F1,0x1362,0x0351,0x0361,0x0371,0x1372,/*9200*/
	0x13D1,0x13E1,0x13F1,0x1362,0x0351,0x0361,0x0371,0x1372,/*9400*/
	0x13D1,0x13E1,0x13F1,0x1362,0x0351,0x0361,0x0371,0x1372,/*9600*/
	0x13D1,0x13E1,0x13F1,0x1362,0x0351,0x0361,0x0371,0x1372,/*9800*/
	0x13D1,0x13E1,0x13F1,0x1362,0x0351,0x0361,0x0371,0x1372,/*9A00*/
	0x13D1,0x13E1,0x13F1,0x1362,0x0351,0x0361,0x0371,0x1372,/*9C00*/
	0x13D1,0x13E1,0x13F1,0x1362,0x0351,0x0361,0x0371,0x1372,/*9E00*/
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,/*A000*/
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,/*A200*/
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,/*A400*/
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,/*A600*/
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,/*A800*/
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,/*AA00*/
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,/*AC00*/
	0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,0x0040,/*AE00*/
	0x11D1,0x11E1,0x11F1,0x11A2,0x0851,0x0861,0x0871,0x11B2,/*B000*/
	0x11D1,0x11E1,0x11F1,0x11A2,0x0851,0x0861,0x0871,0x11B2,/*B200*/
	0x11D1,0x11E1,0x11F1,0x11A2,0x0851,0x0861,0x0871,0x11B2,/*B400*/
	0x11D1,0x11E1,0x11F1,0x11A2,0x0851,0x0861,0x0871,0x11B2,/*B600*/
	0x11D1,0x11E1,0x11F1,0x11A2,0x0851,0x0861,0x0871,0x11B2,/*B800*/
	0x11D1,0x11E1,0x11F1,0x11A2,0x0851,0x0861,0x0871,0x11B2,/*BA00*/
	0x11D1,0x11E1,0x11F1,0x11A2,0x0851,0x0861,0x0871,0x11B2,/*BC00*/
	0x11D1,0x11E1,0x11F1,0x11A2,0x0851,0x0861,0x0871,0x11B2,/*BE00*/
	0x15D1,0x15E1,0x15F1,0x08A1,0x0381,0x0141,0x0142,0x08E1,/*C000*/
	0x15D1,0x15E1,0x15F1,0x08A1,0x0381,0x0141,0x0142,0x08E1,/*C200*/
	0x15D1,0x15E1,0x15F1,0x08A1,0x0381,0x0141,0x0142,0x08E1,/*C400*/
	0x15D1,0x15E1,0x15F1,0x08A1,0x0381,0x0141,0x0142,0x08E1,/*C600*/
	0x15D1,0x15E1,0x15F1,0x08A1,0x0381,0x0141,0x0142,0x08E1,/*C800*/
	0x15D1,0x15E1,0x15F1,0x08A1,0x0381,0x0141,0x0142,0x08E1,/*CA00*/
	0x15D1,0x15E1,0x15F1,0x08A1,0x0381,0x0141,0x0142,0x08E1,/*CC00*/
	0x15D1,0x15E1,0x15F1,0x08A1,0x0381,0x0142,0x0142,0x08E1,/*CE00*/
	0x1391,0x13A1,0x13B1,0x1322,0x0311,0x0321,0x0331,0x1332,/*D000*/
	0x1391,0x13A1,0x13B1,0x1322,0x0311,0x0321,0x0331,0x1332,/*D200*/
	0x1391,0x13A1,0x13B1,0x1322,0x0311,0x0321,0x0331,0x1332,/*D400*/
	0x1391,0x13A1,0x13B1,0x1322,0x0311,0x0321,0x0331,0x1332,/*D600*/
	0x1391,0x13A1,0x13B1,0x1322,0x0311,0x0321,0x0331,0x1332,/*D800*/
	0x1391,0x13A1,0x13B1,0x1322,0x0311,0x0321,0x0331,0x1332,/*DA00*/
	0x1391,0x13A1,0x13B1,0x1322,0x0311,0x0321,0x0331,0x1332,/*DC00*/
	0x1391,0x13A1,0x13B1,0x1322,0x0311,0x0321,0x0331,0x1332,/*DE00*/
	0xC410,0xC420,0xC430,0x0400,0xC450,0xC460,0xC470,0x0440,/*E000*/
	0xC410,0xC420,0xC430,0x0480,0xC450,0xC460,0xC470,0x04C0,/*E200*/
	0xC410,0xC420,0xC430,0x0500,0xC450,0xC460,0xC470,0x0540,/*E400*/
	0xC410,0xC420,0xC430,0x0580,0xC450,0xC460,0xC470,0x05C0,/*E600*/
	0xC410,0xC420,0xC430,0x160D,0xC450,0xC460,0xC470,0x364D,/*E800*/
	0xC410,0xC420,0xC430,0x168D,0xC450,0xC460,0xC470,0x36CD,/*EA00*/
	0xC410,0xC420,0xC430,0x170D,0xC450,0xC460,0xC470,0x374D,/*EC00*/
	0xC410,0xC420,0xC430,0x178D,0xC450,0xC460,0xC470,0x37CD,/*EE00*/
	0x0E00,0x0EC0,0x6E80,0x6EB0,0x0F00,0x0F40,000000,000000,/*F000*/
	0x0C00,0x0CC0,0x0C40,0x6CB0,0x0D00,0x0D40,000000,000000,/*F200*/
	0x0D80,0x0D80,0x0D80,0x0D80,0x0E40,0x1400,0x1440,0x1440,/*F400*/
	0x0BC4,000000,000000,000000,000000,000000,000000,000000,/*F600*/
	0x0DC0							/*F800*/
};

static const char oz_imz[] = {2,2,2,4,4,8,12,12,0};

static const char oz_char[] = "bwlsdxp";

static const char fpozx[] =
	{_oz_l,_oz_s,_oz_x,_oz_p,_oz_w,_oz_d,_oz_b,_oz_c};

static const char am_size[_am_count] =
{ /* default sizes only; size must be defined for manual */
	0,0,0,0,/* implied;Dn;An;(An) */
	0,0,2,2,/* (An)+;-(An);W(An),<An+ix/cm> */
	2,4,2,2,/* AW;AL;*+W;<*+ix/cm> */
	2,2,0,0	/* #oz;{};CCR;SR */
};

static const char op_size_complex[128] =
{ /* r=(uint8)((v>>4&3)-1),r+=v<<4&v&64; if(v&=7) r+=(uint8)((v&3)-1); return r>4?-1:r+r+2; */
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	 2, 2, 4, 6,-1, 2, 4, 6, 2, 2, 4, 6,-1, 2, 4, 6,
	 4, 4, 6, 8,-1, 4, 6, 8, 4, 4, 6, 8,-1, 4, 6, 8,
	 6, 6, 8,10,-1, 6, 8,10, 6, 6, 8,10,-1, 6, 8,10,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	 2, 2, 4, 6,-1,-1,-1,-1, 2, 2, 4, 6,-1,-1,-1,-1,
	 4, 4, 6, 8,-1,-1,-1,-1, 4, 4, 6, 8,-1,-1,-1,-1,
	 6, 6, 8,10,-1,-1,-1,-1, 6, 6, 8,10,-1,-1,-1,-1
};

/* returns instruction size and op type */
int disasm68K(const uint8 *code,
	int offs,
	int size,
	uint8 *inst,
	uint8 *oper,
	uint8 *comm)
{
	const uint8	*s;
	int		z, o, sofs, dofs;
	int		dm, sm = 0, mm; /* dst, src and modifier modes */
	int		t, oz, imsgn;
	uint32		op, ea;

	if (offs+0u>=size)
		return 0;
	s = code+offs;
	op = be16_native(*(const uint16*)s);
	o = opinfo[op>>6];
	z = am_size[dm=o&15]+2;
	ea = op&63;
	oz = o>>4&3;
	mm+= _am_sr&-((mm=o>>7+6)!=0);
	dofs = sofs = 2, imsgn = _am_imz;
	switch(o=o>>6&127)
	{
	case _op_swap:
		if ((uint8)ea>15)
			{o = _op_pea; goto fea;}
		if ((uint8)ea>7)
			{o = _op_bkpt, op&= 7, mm = _am_sq; break;}
		goto gea;
	case _op_ext:
		if ((uint8)ea>7)
		{ /* MOVEM <list>,<mea> */
			z = 4, o = _op_movem, mm = _am_rnl;
		case _op_psave:
		case _op_fsave:
			goto noaip;
		}
		goto gea; /* all other ea values accounted for */
	case _op_movem: /* MOVEM <mea>,list */
		dm = _am_rnl; /* field in opinfo */
	case _op_prestore:
	case _op_frestore:
		if (ea-32u<8)
			goto bad; /* -(An) not allowed for src */
		goto mea;
	case _op_exg:
		if ((uint8)ea>15)
			{o = _op_and; goto gea;}
		if ((char)op<(char)0x88)
			goto bad;
		/* not completely necessary to use mmode */
		/* but matching for completeness; left operand is op0 Rh */
		sm = dm, dm = 0, mm = _am_dn;
		if ((uint8)op-0x48u<8)
			sm = mm = _am_an;
	case _op_moveq: /* mm/dm already set */
		break;
	case _op_extb:
		if ((uint8)ea>7 || !dm)
			{o = _op_lea, dm = _am_an, oz = 0; goto fea;}
		op<<= 9;
		break;
	case _op_ori:
	case _op_andi:
	case _op_eori:
		if ((uint8)ea==60)
		{ /* CCR/SR destination */
			dm = oz-_oz_b;
			if ((uint8)dm+0u>1)
				goto bad;
			dm+= _am_ccr, oz = 0;
			break;
		}
	case _op_subi:
	case _op_addi:
	case _op_cmpi:
		z+= (oz-_oz_b)&2; /* account for L */
		goto nea;
	case _op_btst:
	case _op_bchg:
	case _op_bclr:
	case _op_bset:
		if ((uint8)dm==_am_imz) /* static; bit index imm byte (only low 3 bits used) */
			goto dea; /* whole word exposed for debugging purposes */
		if (ea-8u>7) /* dynamic shift=1st operand */
			{mm = _am_dn, dm = 0; goto dea;}
		o = _op_movep, z = 4;
		oz = (op>>6&1)+_oz_w;
		sm = dm = _am_dn;
		if ((int8)op<0)
			dm = _am_dai;
		else
			sm = _am_dai;
		break;
	case _op_asr:
	case _op_asl:
	case _op_lsr:
	case _op_lsl:
	case _op_roxr:
	case _op_roxl:
	case _op_ror:
	case _op_rol:
		if (!mm) /* ea destination, 1-bit shift */
			goto mea; /* this operation is memory-only */
		o+= (ea>>3-1)&(3<<1);
		sm = _am_dn; /* dest reg. is in low 3 bits */
		if ((uint8)ea>31)
			mm = _am_dn;
		break;
	case _op_eor:
		if (ea-8u<8)
			{o = _op_cmpm, sm = dm = _am_aip; goto gea;}
		mm = _am_dn, dm = 0;
		goto dea; /* Dn,<ea> only */
	case _op_pack:
	case _op_unpk:
		if ((uint8)ea>15)
			{o = _op_or; goto nea;} /* <ea>,Dn */
		z = 4, mm = _am_imz;
		goto dnmai;
	case _op_sbcd: /* byte operation */
		if ((uint8)ea>15)
			{o = _op_or, mm = dm, dm = 0; goto dea;} /* Dn,<ea> */
		goto dnmai;
	case _op_addx:
		if ((uint8)ea>15)
			{o = _op_add, mm = dm, dm = 0; goto dea;} /* Dn,<ea> */
		goto dnmai;
	case _op_subx:
		if ((uint8)ea>15)
			{o = _op_sub, mm = dm, dm = 0; goto dea;} /* Dn,<ea> */
		goto dnmai;
	case _op_abcd: /* byte operation */
		if ((uint8)ea>15)
			{o = _op_and, mm = dm, dm = 0; goto dea;} /* Dn,<ea> */
	dnmai:	sm = _am_dn;
		if ((uint8)ea>7)
			sm = dm = _am_mai;
		break;
	case _op_nbcd:
		if (ea-8u<8)
			{o = _op_link, z = 6, oz = _oz_l, mm = _am_isz; break;}
		goto dea;
	case _op_divsl: /* MULx.L <ea>,Dn[:Dl] */
		dm = _am_dh, o = _op_mulu;
		t = s[2];
		if ((uint8)(t&4))
			dm = _am_dlh;
		if ((uint8)(t&8))
			{o+= _op_muls-_op_mulu; goto mds;}
		goto nea;
	case _op_divul:
		dm = _am_dlh;
		t = s[2];
		if ((uint8)(t&4))
			o = _op_divu;
		else if (!(((t>>4)^s[3])&7))
			dm = _am_dh, o = _op_divu;
		if ((uint8)(t&8))
		{
			o+= _op_divs-_op_divu;
		case _op_muls:
		case _op_divs:
		mds:	imsgn = _am_isz;
			goto nea;
		}
		goto nea;
	case _op_cas:
		if ((uint8)ea==60)
		{ /* CAS2 Dc1:Dc2,Du1:Du2,(Rn1):(Rn2) */
			if ((uint8)oz<_oz_w)
				goto bad;
			z = 6, o = _op_cas2;
			mm = _am_dlc, sm = _am_dmu, dm = _am_rnc;
			break;
		} /* all non-ea operands in mm due to sm ordering */
		dm = 0; /* Dl0,Dm6,<mea> */
		goto mea;
	case _op_cas2: /* opinfo placeholder for MOVE CCR/SR,ea */
		o = _op_move, mm = dm, dm = 0;
	dea:	if ((uint8)ea>57) /* destination data <ea>; no imms/rels allowed */
			goto bad;
		goto nea;
	case _op_bra:
	case _op_bsr:
	case _op_bcc:
		if (!(char)op)
			z = 4, oz = _oz_w+256;
		else if ((char)op==-1)
			z = 6, oz = _oz_l;
		break;
	case _op_scc:
		if ((uint8)ea>57)
		{
			o = _op_trapcc;
			if (!(ea-=58))
				{oz = _oz_w, dm = _am_imz; goto dm_;}
			if (!--ea)
				{oz = _oz_l, dm = _am_imz; goto dm_;}
			o = _op_tcc;
			if (--ea)
				goto bad;
			break;
		}
		if (ea-8u<8)
			{o+= _op_dbcc-_op_scc, sm = _am_dn, dm = _am_rel; goto dm_;}
		goto gea; /* all other ea values accounted for */
	case _op_trap: /* TRAP/RTE/TRAPV/RTR.. */
		switch(ea)
		{
		case 0: case 1: case 2: case 3:
		case 4: case 5: case 6: case 7:
		case 8: case 9: case 10: case 11:
		case 12: case 13: case 14: case 15:
			mm = _am_sq, op&= 15;
			break;
		case 16: case 17: case 18: case 19:
		case 20: case 21: case 22: case 23:
			o = _op_link, sm = _am_an, dm = _am_isz;
			goto dm_;
		case 24: case 25: case 26: case 27:
		case 28: case 29: case 30: case 31:
			o = _op_unlk, sm = _am_an;
			break;
		case 32: case 33: case 34: case 35: /* MOVE Ax,USP|.. */
		case 36: case 37: case 38: case 39:
			o = _op_move, sm = _am_an, dm = _am_usp;
			break;
		case 40: case 41: case 42: case 43: /* MOVE ..|USP,Ax */
		case 44: case 45: case 46: case 47:
			o = _op_move, sm = _am_usp, dm = _am_an, op<<= 9;
			break;
		case 48: o = _op_reset; break;
		case 49: o = _op_nop; break;
		case 50: o = _op_stop, dm = _am_imz; goto dm_; /* #->SR */
		case 51: o = _op_rte; break;
		case 52: o = _op_rtd, dm = _am_isz; goto dm_; /* #W */
		case 53: o = _op_rts; break;
		case 54: o = _op_trapv; break;
		case 55: o = _op_rtr; break;

		case 58:
		case 59:
			o = _op_movec, z = 4;
			t = be16_native(((const uint16*)s)[1]);
			if ((uint16)(t&0x7F0))
				goto bad;
			t&= 0xFFF;
			if ((uint8)(t&8))
				if ((uint16)t==8)
					dm = _am_buscr;
				else if ((uint16)t==0x808)
					dm = _am_pcr;
				else
					goto bad;
			else
				dm = (t>>8)+(uint8)t+_am_sfc;
			sm = dm;
			if ((char)(op&1))
				sm = _am_rn;
			else
				dm = _am_rn;
			break;

		default: goto bad;
		}
		break;
	case _op_rtm: /* 68020-only */
		if ((uint8)ea>15)
			o = _op_callm, dm = _am_imz, oz = _oz_b+256; /* CALLM #B,<mea> */
		goto gea; /* RTM Rn */
	case _op_tas:
		if ((uint8)ea==58)
			{o = _op_bgnd; break;}
		goto nea;
	case _op_cpusha: /* CPU32-only instructions */
		z = 4;
		if ((uint16)((t=be16_native(((const uint16*)s)[1]))&0x8238))
			goto bad;
		if ((uint16)t==0x1C0)
		{
			o = _op_lpstop;
			if (ea) /* optional */
				goto bad;
			dm = _am_imz; /* #->SR */
			goto dm_;
		}
		/* TBLU[N]/TBLS[N] */
		o = _op_tblu+(t>>10&3), oz = t>>6&3;
		if (oz>2)
			goto bad;
		oz+= _oz_b;
		dm = _am_dh;
		if ((uint16)(t&256))
		{
		noaip:	if (ea-24u<8)
				goto bad; /* (An)+ not allowed */
			goto mea;
		}
		sm = _am_dmn; /* Dym:Dyn,Dx */
		break;
	case _op_cinva: /* CINVx/CPUSHx [cache,](An) */
		mm = (op>>6&3)+_am_nc;
		o = (ea>>5)+_op_cinva;
		ea&= 31;
		if ((uint8)ea<8)
			goto bad;
		if ((uint8)ea<24)
			sm = _am_ai, o+= (uint8)ea<16 ? _op_cinvl-_op_cinva:_op_cinvp-_op_cinva;
		break;
	case _op_pmove:
		z = 4, t = be16_native(((const uint16*)s)[1]);
		/* all read/write modes from bit #9 are W=0;R=1 */
		switch(o=t>>10)
		{
			static const char pmt[] =
			{
				_am_tc,_am_drp,_am_srp,_am_crp,
				_am_cal,_am_val,_am_scc,_am_ac,
				_am_psr,_am_pcsr,0,0,
				_am_badx,_am_bacx
				
			};
		/* PMOVE <p-reg.> */
		case 2:	case 3:
			dm = o+_am_tt0-2; /* note that TTx is AXx on EC030 */
			goto mrm;
		case 0: case 1: case 4: case 5: case 6: case 7:	goto bad; /* bad index */
		/* PLOAD <fc>,<mea> */
		case 8: o = _op_ploadw+(t>>9&1);
			goto ptc;
		/* PVALID val|An,<fea> */
		case 10: case 11: mm = _am_val+10-op;
			o = _op_pvalid; /* fixme */
			if ((uint16)(t&0x3F8))
				goto bad;
			goto fea;
		/* PFLUSH <fc>,#mask; bit #0 of mode=shared */
		case 8+1: o = _op_pflusha; break; /* no operands */
		case 8+4: ea = 64; /* (no ea) */
		case 8+6: o = _op_pflush, dm = _am_pfk; goto pfc;
		case 8+5: ea = 64; /* (no ea) */
		case 8+7: o = _op_pflushs;
		pfc:	if ((uint16)(t&0x200))
				goto bad; /* no read mode */
			dm = _am_pfk;
			goto ptc;

		default:
			if ((uint16)t==0xA000)
				{o = _op_pflushr; goto mea;}
			if ((uint16)(t&0xE000)==0x8000)
			{
		/* PTEST FC,<mea>,#<level>[,An] */
				op = ea, op|= t&256;
				o = _op_ptestw+(t>>9&1);
				dm = _am_pta;
			ptc:	t&= 31;
				if ((uint8)(t&16))
					mm = _am_dc3;
				else if ((uint8)(t&8))
					mm = _am_dn, op = (uint8)op|t<<9;
				else if ((uint8)t<2)
					mm = _am_sfc+t;
				else
					goto bad;
				if (ea-24u<16)
					goto bad;
				if ((uint8)ea>63)
					break;
				goto fea;
			}
		/* PMOVE MRn,<mea> | <mea>,MRn */
			if (o>29 || (dm=pmt[o-16])==0)
				goto bad;
		mrm:	o = _op_pmove;
			if ((uint8)dm<=_am_pcsr && (char)t)
				goto bad;
			if ((uint16)(t&256))
			{
				if (dm-_am_drp+0u>_am_psr-_am_drp)
					goto bad;
				o = _op_pmovefd;
			}
			if ((uint16)(t&512)) /* read mode */
			{
				if ((uint16)o==_op_pmovefd)
					goto bad;
				mm = dm, dm = 0;
			} else if ((uint8)dm==_am_pcsr)
				goto bad;
			goto fea;
		}
	case _op_pflush: /* '040 */
		sm = _am_ai;
		if ((uint8)ea>15)
		{
			sm = 0;
			if ((uint8)ea>31)
				goto bad;
			o = _op_pflushan;
			if ((uint8)ea>23)
				o = _op_pflusha;
		} else if ((uint8)ea<8)
			o = _op_pflushn;
		break;
	case _op_chk2: /* PTEST - '040 */
		o = _op_ptestw+(ea>>5&1);
		goto alai;
	case _op_cmpm: /* PLPA - '060 */
		o = _op_plpaw+(op>>6&1);
	alai:	sm = _am_ai;
		if ((uint8)(ea&24)!=8)
			goto bad;
		break;
	case _op_pscc:
		z = 4;
		if (be16_native(((const uint16*)s)[1])>31)
			goto bad;
		if (ea-8u<7)
			{o = _op_pdbcc, sm = _am_dn, dm = _am_rel; goto dm_;}
		if ((uint8)ea>57)
			{o = _op_ptrapcc; goto ftz;}
		goto nea; /* 60 already used above */
	case _op_pbcc:
	case _op_fbcc:
		if (oz)
			z = 6;
		else
			z = 4, oz = 256+_oz_w;
		break;
	case _op_fscc:
		z = 4;
		if (ea-8u<7)
			{o = _op_fdbcc, sm = _am_dn, dm = _am_rel; goto dm_;}
		if ((uint8)ea>57)
		{
			o = _op_ftrapcc;
		ftz:	if (!(ea-=58))
				{oz = _oz_w, dm = _am_imz; goto dm_;}
			if (!--ea)
				{oz = _oz_l, dm = _am_imz; goto dm_;}
			if (--ea)
				goto bad;
			break;
		}
		goto nea; /* 60 already used above */
	case _op_fnop:
		if (((const int16*)s)[1])
			{o = _op_fbcc, dm = _am_rel; goto dm_;}
		z = 4; /* functions as a no-op in FBcc.w *+2 form */
		break;
	case _op_fmovem: /* Fxx */
		z = 4;
	/* NOTE: some fmoves have a k-factor, static for P and dynamic instead of .C */
		t = ((const uint8*)s)[2];
		switch(t>>5)
		{
		case 0: /* z is src reg.index, implied .X mode */
			sm = _am_fph;
			goto fop;
		case 2: oz = fpozx[(t>>2)-(2<<5-2)]; /* Fxx.z <ea>,FPn */
			dm = _am_fpn;
			if (oz==_oz_c)
				{o = _op_fmovecr, sm = _am_isz; goto end;}
		fop:	if ((o=fpopx[s[3]&63])==0)
				goto bad;
			if ((uint16)o==_op_fsincos)
				mm = _am_fpl;
			if (sm)
				goto end;
			break;
		case 4: dm = _am_fcl;
			if (((const uint16*)s)[1]&be16_native((1<<10)-1))
				goto bad;
		case 5: mm = _am_fcl;
			if (((const uint16*)s)[1]&be16_native((1<<10)-1))
				goto bad;
			break;
		case 6: dm = (uint8)(t&8) ? _am_fnd:_am_fnl; goto fmv;
		case 7: mm = (uint8)(t&8) ? _am_fnd:_am_fnl;
		fmv:	op = ea;
			if ((uint8)(t&16))
			{
				if (ea-32u<8)
					goto bad; /* -(An) not allowed in postincr. mode */
			} else if (ea-24u<8)
				goto bad; /* (An)+ not allowed in predecr. mode */
			break;
		case 1: case 3: /* there is no 0x1xxxxx pattern */
		default: goto bad;
		}
		imsgn = _am_isz; /* ints with FP ops are always signed */
	default:
	nea:	if (ea-8u<8) /* number/no An general effective address */
			goto bad;
		goto gea;
	case _op_move:
		z = 2; /* remove suggested size as dst codes occur after <ea> codes */
		goto gea;
	case _op_moves: /* <mea>,Rn */
		dm = _am_rn;
		goto mea;
	case _op_cmp2: /* <fea>,Rn */
		o+= s[2]&8; /* _op_chk2 if bit #11 of op1; both are Rn */
		dm = _am_rn;
		//z = 4; //taken care of in opinfo
	case _op_jsr:
	case _op_jmp:
	fea:	if (ea-24u<16) /* fixed/no incr. memory-only <ea> */
			goto bad;
	mea:	if ((uint8)ea<16) /* memory-only <ea> */
			goto bad;
	case _op_addq:
	case _op_subq:
		if ((uint8)ea==60)
			goto bad;
	case _op_movea:
	case _op_adda:
	case _op_suba:
	case _op_cmpa:
	case _op_cmp:
	case _op_add:
	case _op_sub:
	case _op_tst: /* takes ALL general addressing modes on '020+ including imm */
	gea:	o|= ea<<8;
		sofs = z; /* general <ea> */
		switch(ea)
		{
		case 0: case 1: case 2: case 3:
		case 4: case 5: case 6: case 7: sm = _am_dn; break;
		case 8: case 9: case 10: case 11:
		case 12: case 13: case 14: case 15: sm = _am_an; break;
		case 16: case 17: case 18: case 19:
		case 20: case 21: case 22: case 23: sm = _am_ai; break;
		case 24: case 25: case 26: case 27:
		case 28: case 29: case 30: case 31: sm = _am_aip; break;
		case 32: case 33: case 34: case 35:
		case 36: case 37: case 38: case 39: sm = _am_mai; break;
		case 40: case 41: case 42: case 43:
		case 44: case 45: case 46: case 47: z+= 2, sm = _am_dai; break;
		case 59: sm = _am_rex; goto cplx;
		case 48: case 49: case 50: case 51:
		case 52: case 53: case 54: case 55:
			sm = _am_aix;
		cplx:	t = 2;
			if (s[z]&1 && (t=op_size_complex[s[z+1]&127])<0)
			case 61: case 62: case 63:
				goto bad;
			z+= t;
			break;
		case 56: z+= 2, sm = _am_abw; break;
		case 57: z+= 4, sm = _am_abl; break;
		case 58: z+= 2, sm = _am_rel; break;
		case 60: z+= oz_imz[(uint8)oz], sm = imsgn;
		}
		if ((uint8)(dm|(_am_isz-_am_imz))==_am_isz)
			break; /* size already taken care of */
		if ((uint8)o==_op_move) /* uint8 cast is necessary due to OR op above */
		dm_:	dofs = z;
		/* finish size with destination mode */
		if ((uint8)dm==_am_aix)
		{
			t = 2;
			if (s[z]&1 && (t=op_size_complex[s[z+1]&127])<0)
				goto bad;
			z+= t;
		} else if ((uint8)(dm|(_am_isz-_am_imz))==_am_isz)
			z+= oz_imz[(uint8)oz];
		else if (dm!=_am_bf)
			z+= am_size[dm];
		break;
	case _op_move16: /* MOVE16 */
		sm = dm;// = _am_aip;
		if ((uint8)ea<32)
		{ /* <->AL */
			z = 6;
			if ((uint8)(ea&16))
				sm = dm = _am_ai;
			if ((uint8)(ea&8))
				{sm = _am_abl, op<<= 9; break;}
			else
				{dm = _am_abl; break;}
		} else if ((uint8)ea<40 && (t=((int8*)s)[2])<0) /* (Ax)+,(Ax)+ */
			{z = 4, op = (uint32)t<<8-3; break;}
		goto bad;
	case _op_atrap:
	case _op_ftrap:
	case _op_itrap:
		goto xbad;
	}
end:	if (offs+z+0u>size)
	{
	bad:	o = (uint16)op>0xEFFF ? _op_ftrap:_op_itrap;
	xbad:	z = 2, oz = sm = dm = 0, mm = _am_dcw;
	}
	if (inst)
	{
		char	*d = strcatop((char*)inst+1,(uint8)o,be32_native(*(const uint32*)s));

		if (oz-_oz_b+0u<_oz_c-_oz_b)
			*d++ = '.', *d++ = oz_char[-_oz_b+oz];
		else if ((uint8)(o-_op_bra+0u)<3 && !oz)
			*((int16*)d)++ = MK2CC('.','s');
		*d = 0;
		*inst = (uint8)(d-(char*)inst-1);
	}
	oz<<= 16;
	z|= (oz|o)<<8;
	if (comm)
	{
		char	*d = (char*)comm+1;

		if ((t=strcomam(d,sm,op|oz,s+sofs,offs))==0
				&& (t=strcomam(d,dm,op|oz,s+dofs,offs))==0)
			t = strcomam(d,mm,op|oz,s+2,offs);
		*comm = (uint8)t;
	}
	if (oper)
	{
		char	*d = (char*)oper+1;
		int8	ord = (uint8)dm==_am_imz; /* immediate dmode comes before sm */

		if (mm && ((uint8)o==_op_bfins || ((uint8)mm!=_am_fpl
				&& (uint8)(mm|1)!=_am_rn && (uint8)mm!=_am_imz)))
		{
			d = strcatam(d,mm,((uint8)mm<=_am_aix ? op>>9:op)|oz,s+2), o = 1;
			mm = 0;
		} else
			o = 0;
		t = 2;
		do
		{
			if (!(char)((t^ord)&1) && sm) /* even first with 2 */
			{
				if (o)
					*d++ = ',';
				d = strcatam(d,sm,op|oz,s+sofs), o = 1;
				if ((uint8)dm==_am_bf)
				{
					d = strcatam(d,dm,0,s+2), o = 1; /* ,Dn in mm */
					break;
				}
			}
			if ((char)((t^ord)&1) && dm)
			{
				if (o)
					*d++ = ',';
				d = strcatam(d,dm,((uint8)dm<=_am_aix ? op>>9:op)|oz,s+dofs);
				o = 1;
			}
		} while(--t);
		if (mm)
		{
			if (o)
				*d++ = mm==_am_fpl ? ':':',';
			d = strcatam(d,mm,((uint8)mm<=_am_aix ? op>>9:op)|oz,s+2);
		}
		*d = 0;
		*oper = (uint8)(d-(char*)oper-1);
	}
	return z;
}

static int validMBName(const uint8 *s,
	const uint8 *e)
{
	uint8		c;

	while(s<e)
		if (!isalnum(c=*s++) && c!=' ' && c!='_')
			return 0;
	return 1;
}

int unpackMacsBugCppName(char *d,
	int maxnamelen,
	const char *s)
{
	const char	*n = (const char*)s, *ed = d+maxnamelen, *es;
	int		i;
	uint8		t, fc, c, u, r, p, v, wb, pb, clip;

	if (maxnamelen<=0)
		return 0;
	es = s+strlen(s), u = c = fc = 0;
	for(n=s; s<es; ++s)
	{ /* scan first name */
		if (isalnum(t=*(const uint8*)s) || t==' ' || t=='%')
			continue;
		if (t!='_')
			break;
		if (s<es-2 && *(const int16*)s=='__')
			if (isdigit(t=((const uint8*)s)[2]) || t=='F' || t=='C')
				break;
	}
	i = s-n;
	if (i==4 && *(const int16*)n=='__')
	{ /* check for 'create', 'destroy', operators... */
		switch(((const int16*)n)[1])
		{
		case MK2CC('d','t'): c = 1; break;
		case MK2CC('c','t'): c = 2; break;
		case MK2CC('n','w'): n = " new"; goto op1;
		case MK2CC('d','l'): n = " delete"; goto op1;
		case MK2CC('p','l'): n = "+"; goto op1;
		case MK2CC('m','i'): n = "-"; goto op1;
		case MK2CC('m','l'): n = "*"; goto op1;
		case MK2CC('d','v'): n = "/"; goto op1;
		case MK2CC('m','d'): n = "%"; goto op1;
		case MK2CC('l','s'): n = "<<"; goto op1;
		case MK2CC('r','s'): n = ">>"; goto op1;
		case MK2CC('a','d'): n = "&"; goto op1;
		case MK2CC('o','r'): n = "|"; goto op1;
		case MK2CC('e','r'): n = "^"; goto op1;
		case MK2CC('c','o'): n = "~"; goto op1;
		case MK2CC('n','t'): n = "!"; goto op1;
		case MK2CC('a','s'): n = "="; goto op1;
		case MK2CC('l','t'): n = "<"; goto op1;
		case MK2CC('g','t'): n = ">"; goto op1;
		case MK2CC('e','q'): n = "=="; goto op1;
		case MK2CC('n','e'): n = "!="; goto op1;
		case MK2CC('l','e'): n = "<="; goto op1;
		case MK2CC('g','e'): n = ">="; goto op1;
		case MK2CC('a','a'): n = "&&"; goto op1;
		case MK2CC('o','o'): n = "||"; goto op1;
		case MK2CC('p','p'): n = "++"; goto op1;
		case MK2CC('m','m'): n = "--"; goto op1;
		case MK2CC('c','l'): n = "()"; goto op1;
		case MK2CC('v','c'): n = "[]"; goto op1;
		case MK2CC('r','f'): n = "->";
		op1:	i = strlen(n), c = 4;
		}
	} else if (i>254) /* sanity check*/
		return 0;
	u = i;
	if (s+2>es || *((const int16*)s)++!='__')
		return 0;
	for(i=0; s<es && (t=*s-'0')<=9; ++s)
		if ((i=i*10u+t)>250)
			return 0;
	if (c&3)
	{
		n = s;
		if ((u=i)==0)
			return 0; /* need name */
	}
	clip = 1;
	if (i)
	{ /* has class name */
		if (s+i>es)
			return 0;
		if (d>=ed-3)
			goto cend; /* maxnamelen<4?? */
		if (d+i>ed)
			i = ed-d-2;
		do
			*d++ = *s++;
		while(--i);
		if (d>=ed-3)
			goto cend;
		*((int16*)d)++ = '::';
	}
	if (c)
		if (c==4)
		{
			if (d>=ed-11)
				goto cend;
			*((int32*)d)++ = MK4CC('o','p','e','r');
			*((int32*)d)++ = MK4CC('a','t','o','r');
	 	} else if (!--c)
		{
			if (d>=ed-3)
				goto cend;
			*d++ = '~';
		}
	/* copy function name */
	do
	{
		if (d>=ed-3)
			goto cend;
		*d++ = *n++;
	} while(--u);
	if (s>=es)
		goto end; /* class/name-only is acceptable */
	t = *s++;
	if (t=='C' && s<es)
		fc = 1, t = *s++;
	if (t!='F')
		return 0;
	*d++ = '(';
	for(pb=wb=v=p=r=u=c=0; s<es;)
		switch(t=*s++)
		{
		default: return 0;
		case 'C': if (c) return 0; c = 1; break;
		case 'R': if (r) return 0; r = 1; break;
		case 'S': if (u) return 0; u = 2; break;
		case 'P': if (!++p) return 0;
			c+= c; /* if c before p then const after '*' */
			break;
		case 'U': if (u) return 0; u = 1; break;
		case 'V': if (v) return 0; v = 1; break;
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			for(i=t-'0'; s<es && (t=*s-'0')<=9; ++s)
				if ((i=i*10u+t)>199)
					return 0;
			n = s, s+= i;
			goto tx;
		case 'L': n = "long long"; goto tn; /* new form */
		case 'b': n = "bool"; goto tn; /* new form */
		case 'c': n = "char"; goto tn;
		case 'f': n = "float"; goto tn;
		case 'd': n = "double"; goto tn;
		case 'e': n = "..."; goto tn;
		case 'i': n = "int"; goto tn;
		//case 'j': n = "_stdcall"; goto tn; /* new form */
		//case 'k': n = "class"; goto tn; /* new form */
		case 'l': n = "long"; goto tn;
		case 'm': n = "comp"; goto tn;
		case 'p': n = "extended_80"; goto tn;
		case 'r': n = "long double"; goto tn;
		case 's': n = "short"; goto tn;
		//case 't': n = "template"; goto tn; /* new form */
		case 'v': n = "void"; goto tn;
		case 'w': n = "wchar_t"; goto tn; /* new form */
		case 'x': n = "extended";
		tn:	i = strlen(n);
		tx:	if (d>=ed-4)
				goto cend;
			if (pb)
				*((int16*)d)++ = MK2CC(',',' ');
			if (v)
			{
				if (d>=ed-10)
					goto cend;
				*((int32*)d)++ = MK4CC('v','o','l','a');
				*((int32*)d)++ = MK4CC('t','i','l','e');
				wb = 1;
			}
			if (c==1)
			{
				if (d>=ed-7)
					goto cend;
				*((int32*)d)++ = MK4CC('c','o','n','s'), *d++ = 't';
				wb = 1;
			}
			if (d>=ed-3)
				goto cend;
			if (u)
			{
				if (wb)
					*d++ = ' ';
				if (d>=ed-10)
					goto cend;
				if (u==1)
					*((int16*)d)++ = MK2CC('u','n');
				*((int32*)d)++ = MK4CC('s','i','g','n');
				*((int16*)d)++ = MK2CC('e','d');
				wb = 1;
			}
			if (wb)
				*d++ = ' ';
			if (d+i>=ed-2 && (i=ed-d-2)<=0)
				goto cend;
			memcpy(d,n,i), d+= i;
			wb = 1;
			if (p)
			{
				do
				{
					if (d>=ed-3)
						goto cend;
					*d++ = '*';
				} while(--p);
				wb = 0;
			}
			if (c==2)
			{
				if (d>=ed-8)
					goto cend;
				if (wb)
					*d++ = ' ';
				*((int32*)d)++ = MK4CC('c','o','n','s'), *d++ = 't';
			}
			if (d>=ed-3)
				goto cend;
			if (r)
				*d++ = '&';
			pb = 1, wb = v = p = r = u = c = 0;
		}
	if (c|r|u|p|v) /* all modifiers must be used and cleared */
		return 0;
	*d++ = ')';
	if (fc)
	{
		if (d>=ed-6)
		{
		cend:	if (ed>(n=d+2))
				ed = n;
			while(d<ed)
				*d++ = '.';
			goto end;
		}
		*((int32*)d)++ = MK4CC(' ','c','o','n');
		*((int16*)d)++ = MK2CC('s','t');
	}
end:	*d = 0;
	return 1;
}

int parseMacsBugModuleName(const uint8 *code,
	int z,
	uint16 optype,
	char *name,
	int maxnamelen,
	char *pren)
{
	const uint8	*s = code, *e;
	char		*d;
	int8		typ, clip;

	if ((optype|1)!=_op_rts && optype!=(_op_jmp|16<<8) || z<=0)
		return 0;
	e = s+z;
	if ((z=*s-128)<0)
	{ /* could be 8-char */
		if (s+8>e || (*(const int32*)s==MK4CC('N','u','N','u')
				&& ((const int32*)s)[1]==MK4CC('N','u','N','u')))
			return 0;
		typ = 0, z = 8; /* space-padded 8 byte symbol; no trailer */
	} else
	{
		++s, typ = 2;
		if (z)
		{
			typ = 1; /* length in first byte */
			if (z>31)
			{ /* first char encoded in first byte */
				if (name)
					*name = z, ++name;
				typ = 1|4, z = 7, --maxnamelen;
			}
		} else if ((z=*s++)==0) /* length in second */
			return 0;
		if (s+z>e)
			return 0;
	}
	/* don't print if string is not valid */
	if (!validMBName(s,s+z))
		return 0;
	if (name && maxnamelen>0)
	{
		int		i = z;

		clip = 0;
		if (i>maxnamelen)
			i = maxnamelen-1, clip = 1+(maxnamelen>1);
		memcpy(name,s,i);
		if (clip)
			*(int16*)(name+i-clip) = '..';
		name[i] = 0;
	}
	if ((d=pren)!=0)
	{
		if (typ)
		{
			*((int32*)d)++ = typ&2 ? MK4CC('$','8','0',',')
				:MK4CC('$','8','0','+');
			if (z>9)
			{
				*d++ = '$';
				if (z>15)
					*d++ = ubasechars[z>>4];
			}
			if (typ>4)
				*d++ = '\'', *d++ = *code-128, *d++ = '\'';
			else
				*d++ = ubasechars[z&15];
			*d++ = ',';
		}
		*d = 0;
	}
	if ((typ-1u)<4-1)
	{ /* has trailer */
		s+= z, s+= 2+((int)s&1);
		if (s<=e)
		{
			int		u = be16_native(*(const uint16*)(s-2));

			z|= (MIN(u,65532)+2)<<16;
		}
	}
	return typ<<8|z;
}
